unit Partition;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Native;

type

{$A-,Z+}

 TPartEntry = record
	boot_ind    : Byte;		// 0x80 - active
	head        : Byte;		// starting head
	sector      : Byte;		// starting sector
	cyl         : Byte;		// starting cylinder
	sys_ind     : Byte;		// What partition type
	end_head    : Byte;		// end head
	end_sector  : Byte;	   // end sector
	end_cyl     : Byte;		// end cylinder
	start_sect  : UINT;	   // starting sector counting from 0
	nr_sects    : UINT;  	// nr of sectors in partition
  end;

  TDiskTable = record
   Data  : array[1..446] of Char;
   Table : array[1..4] of TPartEntry;

   Signature : WORD;
  end;

{$A+,Z+}

 TPartition = class
// private
  public
   DriveNo              : Integer;
   PartNo               : Integer;
   FileName             : String;
   DeviceName           : String;

   Cylinders         : _LARGE_INTEGER;
   MediaType         : Integer;
   TracksPerCylinder : DWORD;
   SectorsPerTrack   : DWORD;
   BytesPerSector    : DWORD;

   PartitionType        : Integer;
   BootIndicator        : BOOLEAN;
   RecognizedPartition  : BOOLEAN;
   RewritePartition     : BOOLEAN;
   StartingOffset       : _LARGE_INTEGER;//TLargeInteger;
   PartitionLength      : _LARGE_INTEGER;//TLargeInteger;
   HiddenSectors        : DWORD;

 public
   procedure Dump;
 end;

 TPartitionList = class
 public
   ScanFloppy : Boolean;

   ReiserFSCount : Integer;

   constructor Create;
   destructor Destroy; override;

   procedure NT_Scan;
   procedure LL_Scan(PartType : Integer);
   procedure Win95_Scan(PartType : Integer; UseEI13 : Boolean);
   procedure Dump;
   function Count : Integer;
   function GetPartition(i : Integer) : TPartition;

 private
   Partitions : TList;

   function MediaPresent(h : THandle) : Boolean;
 end;


 function YesNo(b : Boolean) : String;


implementation

uses WinIOCTL, ex2explore, DiskIO, ex2, lvm2;

function StartsWith(S : String; Start : String; var Value : String) : Boolean;
var
   p : Integer;
begin
   p := Pos(Start, S);
   if p = 1 then
   begin
      Result := True;
      Value := Copy(S, p + Length(Start), Length(S));
   end
   else
   begin
      Result := False;
   end;
end;

function EndsWith(S : String; Ends : String; var Value : String) : Boolean;
begin
   if Copy(S, Length(S) - Length(Ends) + 1, Length(Ends)) = Ends then
   begin
      Result := True;
      Value := Copy(S, 1, Length(S) - Length(Ends));
   end
   else
   begin
      Result := False;
   end;
end;
function YesNo(b : Boolean) : String;
begin
   if b then
   begin
      Result := 'Yes';
   end
   else
   begin
      Result := 'No';
   end;
end;

function EnableDASD(h : THandle) : Boolean;
var
   ReturnedByteCount : DWORD;
begin
   Result := DeviceIoControl(h,
                CtlCode(FILE_DEVICE_FILE_SYSTEM, 32, METHOD_NEITHER, FILE_ANY_ACCESS),
                nil,
                0,
                nil,
                0,
                ReturnedByteCount,
                nil);
                if Result then
                begin
                  Debug('DASD Enabled', DebugHigh);
                end
                else
                begin
                  Debug('DASD could not be enabled', DebugHigh);
                end;
end;

function LockVolume(h : THandle) : Boolean;
var
   ReturnedByteCount : DWORD;
begin
   Result := DeviceIoControl(h,
                CtlCode(FILE_DEVICE_FILE_SYSTEM, 6, METHOD_BUFFERED, FILE_ANY_ACCESS),
                nil,
                0,
                nil,
                0,
                ReturnedByteCount,
                nil);
   if Result then
   begin
      Debug('Locking volume: Success', DebugHigh);
   end
   else
   begin
      Debug('Locking volume: Failed', DebugHigh);
   end;
end;

function UnLockVolume(h : THandle) : Boolean;
var
   ReturnedByteCount : DWORD;
begin
   Result := DeviceIoControl(h,
                CtlCode(FILE_DEVICE_FILE_SYSTEM, 7, METHOD_BUFFERED, FILE_ANY_ACCESS),
                nil,
                0,
                nil,
                0,
                ReturnedByteCount,
                nil);
end;

constructor TPartitionList.Create;
begin
   ScanFloppy := False;
   Partitions := TList.Create;
   ReiserFSCount := 0;
end;

destructor TPartitionList.Destroy;
var
   i : Integer;
begin
   for i := 0 to Partitions.Count - 1 do
   begin
      TPartition(Partitions[i]).Free;
   end;
   Partitions.Free;
end;

function TPartitionList.Count : Integer;
begin
   Result := Partitions.Count;
end;

function TPartitionList.GetPartition(i : Integer) : TPartition;
begin
   Result := TPartition(Partitions[i]);
end;

// This does a Low Level scan (reads partition table from disk);
procedure TPartitionList.LL_Scan(PartType : Integer);
var
   DriveNo  : Integer;
   FileName : String;
   h        : THandle;
   Err      : Integer;
   Geometry : TDISK_GEOMETRY;
   Length   : DWORD;
   PartNo   : Integer;
   ExtendedStart : ULong;

   procedure AddFloppy(h : THandle);
   var
      Size     : DWORD;
      NewPart  : TPartition;
   begin
      NewPart := TPartition.Create;
      Partitions.add(NewPart);
      NewPart.DriveNo   := DriveNo;
      NewPart.FileName  := FileName;

      NewPart.Cylinders          := Geometry.Cylinders;
      NewPart.TracksPerCylinder  := Geometry.TracksPerCylinder;
      NewPart.SectorsPerTrack    := Geometry.SectorsPerTrack;
      NewPart.BytesPerSector     := Geometry.BytesPerSector;
      NewPart.MediaType          := Geometry.MediaType;

      Size := {Round}(Geometry.Cylinders.QuadPart) * Geometry.TracksPerCylinder * Geometry.SectorsPerTrack;

      NewPart.PartitionType            := PartType;
      NewPart.BootIndicator            := False;
      NewPart.RecognizedPartition      := True;
      NewPart.RewritePartition         := False;
      NewPart.PartNo                   := 0;
      NewPart.StartingOffset.QuadPart  := 0;
      NewPart.PartitionLength.QuadPart := Size;
      NewPart.PartitionLength.QuadPart := NewPart.PartitionLength.QuadPart * Geometry.BytesPerSector;
      NewPart.HiddenSectors            := 0;
   end;

   function ProcessSector(h : THandle; Sector : ULong; Depth : Integer) : Boolean;
   var
      Size     : DWORD;
      Buffer   : TDiskTable;
      i        : Integer;
      SeekPos  : _LARGE_INTEGER; //TLargeInteger;
      NewPart  : TPartition;
   begin
      Result := True;
      Size := sizeof(Buffer);
      Debug('Looking for table at sector ' + IntToStr(Sector), DebugMedium);


      SeekPos.QuadPart := Sector;
      SeekPos.QuadPart := SeekPos.QuadPart * 512;

      if SetFilePointer(h, SeekPos.LowPart, @SeekPos.HighPart, FILE_BEGIN) = $FFFFFFFF then
      begin
         Debug('Seek error ' + IntToStr(GetLastError), DebugOff);
         Result := False;
      end
      else
      begin
         if not ReadFile2(h, @Buffer, Size, Size, nil) then
         begin
            Result := False;
            exit;
         end;
         Debug('Signature is ' + IntToHex(Buffer.Signature, 4), DebugMedium);
         if (Buffer.Signature = $AA55) then
         begin
            for i := 1 to 4 do
            begin
               Debug('Processing partitions', DebugMedium);
               if Buffer.Table[i].sys_ind <> 0 then
               begin
                  Debug('Partition           ' + IntToStr(i) + ' /dev/hdx' + IntToStr(PartNo), DebugMedium);
                  Debug(' boot indicator      ' + IntToHex(Buffer.Table[i].boot_ind,2), DebugHigh);
                  Debug(' starting cylinder   ' + IntToStr(Buffer.Table[i].cyl), DebugHigh);
                  Debug(' starting head       ' + IntToStr(Buffer.Table[i].head), DebugHigh);
                  Debug(' starting sector     ' + IntToStr(Buffer.Table[i].sector), DebugHigh);
                  Debug(' partition type      ' + IntToHex(Buffer.Table[i].sys_ind,2), DebugMedium);
                  case Buffer.Table[i].sys_ind of
                     PARTITION_EXTENDED         : Debug('                     Extended... (dos)', DebugMedium);
                     PARTITION_EXTENDED_LINUX   : Debug('                     Extended... (linux)', DebugMedium);
                     PARTITION_EXTENDED_WIN98   : Debug('                     Extended... (win98)', DebugMedium);
                     PARTITION_FAT_16           : Debug('                     FAT 16', DebugMedium);
                     PARTITION_HUGE             : Debug('                     VFAT', DebugMedium);
                     PARTITION_IFS              : Debug('                     NTFS', DebugMedium);
                     PARTITION_LINUX            : Debug('                     Linux', DebugMedium);
                     PARTITION_LINUX_SWAP       : Debug('                     Linux Swap', DebugMedium);
                  else
                     Debug('Unknown', DebugMedium);
                  end;
                  Debug(' end cylinder        ' + IntToStr(Buffer.Table[i].end_cyl), DebugHigh);
                  Debug(' end head            ' + IntToStr(Buffer.Table[i].end_head), DebugHigh);
                  Debug(' end sector          ' + IntToStr(Buffer.Table[i].end_sector), DebugHigh);
                  Debug(' starting sector     ' + IntToStr(Buffer.Table[i].start_sect), DebugMedium);
                  Debug(' number of sectors   ' + IntToStr(Buffer.Table[i].nr_sects), DebugMedium);
                  Debug('', DebugMedium);


//                  else
                  begin
                     if (Buffer.Table[i].sys_ind = PartType) or (PartType = 0) then
                     begin
                        NewPart := TPartition.Create;
                        Partitions.add(NewPart);
                        NewPart.DriveNo   := DriveNo;
                        NewPart.FileName  := FileName;

                        NewPart.Cylinders          := Geometry.Cylinders;
                        NewPart.TracksPerCylinder  := Geometry.TracksPerCylinder;
                        NewPart.SectorsPerTrack    := Geometry.SectorsPerTrack;
                        NewPart.BytesPerSector     := Geometry.BytesPerSector;
                        NewPart.MediaType          := Geometry.MediaType;

                        NewPart.PartitionType            := Buffer.Table[i].sys_ind;
                        NewPart.BootIndicator            := (Buffer.Table[i].boot_ind = $80);
                        NewPart.RecognizedPartition      := True;
                        NewPart.RewritePartition         := False;
                        NewPart.PartNo                   := PartNo;
                        NewPart.StartingOffset.QuadPart  := (Buffer.Table[i].start_sect + Sector);// Offset from start of partition
                        NewPart.StartingOffset.QuadPart  := NewPart.StartingOffset.QuadPart * Geometry.BytesPerSector;
                        NewPart.PartitionLength.QuadPart := Buffer.Table[i].nr_sects;
                        NewPart.PartitionLength.QuadPart := NewPart.PartitionLength.QuadPart * Geometry.BytesPerSector;
                        NewPart.HiddenSectors            := 0;

                     end;
                     if (Buffer.Table[i].sys_ind in EXTENDED_PARTITIONS) and (PartNo > 4) then
                     begin
                        Dec(PartNo); // We do not want to assign a number to extend partitions after the 1st (I don't know why?, it is what Linux does)
                     end;
                  end;
                  Inc(PartNo);
               end;
            end;
            for i := 1 to 4 do
            begin
               if Buffer.Table[i].sys_ind in EXTENDED_PARTITIONS then
               begin
                  Debug('Processing extended partition....', DebugMedium);
                  if Depth < 32 then
                  begin
                     if PartNo < 5 then
                     begin
                        PartNo := 5;
                     end;
                     if Depth = 1 then
                     begin
                        ExtendedStart := Sector;
                     end;
                     // There is a new partition table at offset start_sect...
                     if Depth >= 2 then
                     begin
                     if not ProcessSector(h, ExtendedStart + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Debug('Did not seem to be a valid extended partition', DebugMedium);
                        end;
                     end
                     else
                     begin
                        if not ProcessSector(h, Sector + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Debug('Did not seem to be a valid extended partition', DebugMedium);
                        end;
                     end;
                  end
                  else
                  begin
                     Debug('More than 32 partitions detected! This is most likely a bug!', DebugOff);
                  end;
               end;
            end;
         end
         else
         begin
            Result := False;
         end;
      end;
   end;
begin
   DriveNo := 0;

   while true do
   begin
      Debug('Opening device...', DebugHigh);
      FileName := '\\.\PHYSICALDRIVE' + IntToStr(DriveNo);
      Debug(FileName, DebugMedium);
      h := Windows.CreateFile(PChar(FileName), GENERIC_READ, FILE_SHARE_READ or FILE_SHARE_WRITE	, nil, OPEN_EXISTING, 0, 0);
      if h = INVALID_HANDLE_VALUE then
      begin
         Err := GetLastError;
         if Err = 5 then
         begin
            MessageDlg('This program requires Administrator privilages to run', mtError, [mbOK], 0);
            break; // out of while loop
         end
         else if Err = 2 then
         begin
            // no more physical drives found...
            break;
         end
         else if Err = 0 then
         begin
            // this is not an error
            break;
         end
         else
         begin
            Debug('Unknown error scanning ' + FileName + ' errorcode = ' + IntToStr(Err), DebugOff);
            Debug(SysErrorMessage(Err), DebugOff);
            break;
         end;
      end
      else // we have the handle, lets look at the partitions....
      begin
         // first see if the media is present...
         if not MediaPresent(h) then
         begin
            Debug('Media not present', DebugOff);
         end
         else
         begin
            Debug('Reading drive geometry...', DebugHigh);
            if not DeviceIOControl(h, CtlCode(FILE_DEVICE_DISK, 0, METHOD_BUFFERED, FILE_ANY_ACCESS),
                              nil, 0,
                              Pointer(@Geometry), sizeof(Geometry),
                              Length, nil) then
            begin
               MessageDlg('Error reading drive geometry. errorcode = ' + IntToStr(GetLastError), mtError, [mbOK], 0);
               break;
            end;
            Debug('Disk Geometry...', DebugHigh);
            Debug(' Cylinders ' + IntToStr(Geometry.Cylinders.QuadPart), DebugHigh);
            Debug(' ' + MediaDescription(Geometry.MediaType), DebugMedium);
            Debug(' Tracks/Cylinder ' + IntToStr(Geometry.TracksPerCylinder), DebugHigh);
            Debug(' Sectors/Track   ' + IntToStr(Geometry.SectorsPerTrack), DebugHigh);
            Debug(' Bytes/Sector    ' + IntToStr(Geometry.BytesPerSector), DebugHigh);
            // Time to read the table
            PartNo := 1;
            ExtendedStart := 0;
            ProcessSector(h, 0, 0);
         end;
         CloseHandle(h);
         Inc(DriveNo);
      end;
   end;

   ////////////////////////////////////////////
   ////////////////////////////////////////////
   ////////////////////////////////////////////
   ////////////////////////////////////////////

   if ScanFloppy then
   begin
      DriveNo := 0;

      while true do
      begin
         FileName := '\\.\' + Char(Ord('A') + DriveNo) + ':';
         Debug(FileName, DebugMedium);
         h := CreateFile(PChar(FileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
//         h := CreateFile(PChar(FileName), 0, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
         if h = INVALID_HANDLE_VALUE then
         begin
            Err := GetLastError;
            if Err = 5 then
            begin
               MessageDlg('This program requires Administrator privilages to run', mtError, [mbOK], 0);
               break; // out of while loop
            end
            else if Err = 2 then
            begin
               // no more physical drives found...
               break;
            end
            else if Err = 21 then
            begin
               Debug('Drive not ready (no disk?)', DebugMedium);
               break;
            end
            else if Err = 0 then
            begin
               // this is not an error
               break;
            end
            else
            begin
               Debug('Unknown error scanning ' + FileName + ' errorcode = ' + IntToStr(Err), DebugOff);
               Debug(SysErrorMessage(Err), DebugOff);
               break;
            end;
         end
         else // we have the handle, lets look at the partitions....
         try
//            LockVolume(h);
            // first see if the media is present...
            begin
               if not DeviceIOControl(h, CtlCode(FILE_DEVICE_DISK, 0, METHOD_BUFFERED, FILE_ANY_ACCESS),
                                 nil, 0,
                                 Pointer(@Geometry), sizeof(Geometry),
                                 Length, nil) then
               begin
                  Err := GetLastError;
                  if Err <> 21 then
                  begin
                     MessageDlg('Error reading drive geometry. errorcode = ' + IntToStr(GetLastError), mtError, [mbOK], 0);
                  end;
                  break;
               end;
               Debug(MediaDescription(Geometry.MediaType), DebugMedium);
               if (Geometry.MediaType = Media_Type_Unknown) and (MediaPresent(h) = False) then
               begin
                  Debug('Media not present', DebugOff);
               end
               else
               begin
                  // Time to read the table
                  PartNo := 1;
                  ExtendedStart := 0;
                  ProcessSector(h, 0, 0);
                  Debug('PartNo = ' + IntToStr(PartNo), DebugMedium);
//                  if PartNo = 1 then
                  begin
                     // no Partitons found, should we add the whole disk?
                     AddFloppy(h);
                  end;
               end;
            end;
         finally
//            UnLockVolume(h);
            CloseHandle(h);
            Inc(DriveNo);
         end;
      end;
   end;
end;

procedure TPartitionList.NT_Scan;
var
   DriveNo    : Integer;
   PartNo     : Integer;
   DeviceName : String;
   h          : THandle;
   Done       : Boolean;
   ErrorNo    : DWORD;
   NewPart    : TPartition;
   i          : Integer;
   NTParts    : TStringList;

   function CheckForReiser(Part : TPartition) : Boolean;
   var
      Seek       : DWORD;
      Buffer     : String;
      Actual     : DWORD;
      TempOffset : _LARGE_INTEGER;
   begin
      Debug('Checking for ReiserFS at ' + IntToStr(Part.StartingOffset.QuadPart), DebugHigh);
      Result := False;
      SetLength(Buffer, 512);
      TempOffset.QuadPart := Part.StartingOffset.QuadPart + 65536;
      Seek := SetFilePointer(h, TempOffset.LowPart, @TempOffset.HighPart, FILE_BEGIN);
      if Seek <> $FFFFFFFF then
      begin
         if ReadFile2(h, PCHAR(Buffer), Length(Buffer), Actual, nil) then
         begin
            // offset 52 should contain 'ReIsEr2Fs'
            if Copy(Buffer, 53, 8) = 'ReIsErFs' then
            begin
               Debug('ReiserFS 3.5 Found. magic = ' + Copy(Buffer, 53, 8), DebugOff);
               Result := True;
            end
            else if Copy(Buffer, 53, 9) = 'ReIsEr2Fs' then
            begin
               Debug('ReiserFS 3.6 Found. magic = ' + Copy(Buffer, 53, 9), DebugOff);
               Result := True;
            end
            else if Copy(Buffer, 53, 9) = 'ReIsEr3Fs' then
            begin
               Debug('ReiserFS JR Found. magic = ' + Copy(Buffer, 53, 9), DebugOff);
               Result := True;
            end
            else if Copy(Buffer, 0, 7) = 'ReIsEr4' then
            begin
               Debug('ReiserFS 4.0 (libreiser) Found. magic = ' + Copy(Buffer, 0, 7), DebugOff);
               Result := True;
            end
            else
            begin
               Debug('No ReiserFS Magic found', DebugHigh);
            end;
         end
         else
         begin
            Debug('ReiserFS Read failed', DebugHigh);
         end;
      end
      else
      begin
         Debug('ReiserFS Seek failed', DebugHigh);
      end
   end;

   function CheckForLVM(Part : TPartition) : Boolean;
   var
      Seek       : DWORD;
      Buffer     : String;
      Magic      : String;
      UID        : String;
      Actual     : DWORD;
      TempOffset : _LARGE_INTEGER;
   begin
      Debug('Checking for LVM2', DebugHigh);
      SetLength(Buffer, 512);

      //LockVolume(h);
      //EnableDASD(h);

     Result := False;

      TempOffset.QuadPart := Part.StartingOffset.QuadPart + 512;
      Seek := SetFilePointer(h, TempOffset.LowPart, @TempOffset.HighPart, FILE_BEGIN);
      Debug('LVM2 Seek result is ' + IntToStr(Seek), DebugHigh);
      if Seek <> $FFFFFFFF then
      begin
         if ReadFile2(h, PCHAR(Buffer), Length(Buffer), Actual, nil) then
         begin
            Debug('LVM2 Read ' + IntToStr(Actual) + ' bytes', DebugHigh);
            //DebugHex(Buffer, DebugHigh);
            // offset 25 should contain 'LVM2'
            Magic := Copy(Buffer, 25, 8);
            Debug('Magic = ''' + Magic + '''', DebugHigh);
            if Magic = 'LVM2 001' then
            begin
               Debug('LVM2 Found. magic = ' + Magic, DebugOff);

               UID := Copy(Buffer, 33, 32);
               Debug('PV UID = ' + UID, DebugOff);
               Result := True;
            end
            else
            begin
               Debug('Magic is not LVM2', DebugHigh);
            end;
         end
         else
         begin
            Debug('LVM2 Read failed ' + IntTostr(GetLastError), DebugHigh);
            //GetDiskExtents(h);
         end;
      end
      else
      begin
         Debug('LVM2 Seek failed', DebugHigh);
      end
   end;

   function TestVolumeRead(Part : TPartition) : Boolean;
   var
      Seek       : DWORD;
      Buffer     : String;
      Actual     : DWORD;
      NewDevice : String;
      NewOffset : Int64;
      NewLength : Int64;
      OffsetObject : TOffset;
   begin
      Debug('Doing test read at' + IntToStr(Part.StartingOffset.QuadPart), DebugHigh);
      SetLength(Buffer, 512);

      Result := False;

      Seek := SetFilePointer(h, Part.StartingOffset.LowPart, @Part.StartingOffset.HighPart, FILE_BEGIN);
      if Seek <> $FFFFFFFF then
      begin
         if ReadFile2(h, PChar(Buffer), Length(Buffer), Actual, nil) then
         begin
            Result := True;
         end;
      end;

      if Result then
      begin
         Debug('Test read was a success', DebugHigh);
      end
      else
      begin
         if Part.StartingOffset.QuadPart = 0 then
         begin
            Debug('Test read failed, trying the back door', DebugHigh);
            if GetDiskExtents(h, NewDevice, NewOffset, NewLength) then
            begin
               Debug('Backdoor is ' + NewDevice, DebugHigh);

               OffsetObject := TOffset.Create;
               OffsetObject.Offset := NewOffset;
               NTParts.AddObject(NewDevice, OffsetObject);

            end;
         end
         else
         begin
            Debug('Test read failed, that''s all I can do.', DebugHigh);
         end;
      end;
   end;

   function CheckForExt(Part : TPartition) : Boolean;
   var
      Seek       : DWORD;
      Superblock : ext2_super_block;
      Actual     : DWORD;
      TempOffset : _LARGE_INTEGER;
   begin
      Debug('Checking for EXT at ' + IntToStr(Part.StartingOffset.QuadPart), DebugHigh);

      Result := False;

      TempOffset.QuadPart := Part.StartingOffset.QuadPart + 1024;
      Seek := SetFilePointer(h, TempOffset.LowPart, @TempOffset.HighPart, FILE_BEGIN);
      if Seek <> $FFFFFFFF then
      begin
         if ReadFile2(h, @superblock, sizeof(superblock), Actual, nil) then
         begin
            if superblock.s_magic = EXT2_SUPER_MAGIC then
            begin
               Debug(DeviceName, DebugLow);
               Debug('EXT2 Found. magic = 0x' + IntToHex(superblock.s_magic, 4), DebugOff);
               Debug('UID = ' + UuidToStr(superblock.s_uuid), DebugOff);
               Result := True;
            end
            else
            begin
               Debug('Magic is not EXT', DebugHigh);
            end;
         end
         else
         begin
            Debug('EXT2 Read failed', DebugHigh);
         end;
      end
      else
      begin
         Debug('EXT Seek failed', DebugHigh);
      end;
   end;

   // Use native directory listings to find devices
   procedure FindNative(AddRemovable : Boolean; PartList : TStringList);
   var
      Devices : TStringList;
      Harddisks : TStringList;
      i : Integer;
      DeviceName : String;
      Number : String;
      DriveNo : Integer;
   begin
      Devices := TStringList.Create;
      Devices.Sorted := True;
      Harddisks := TStringList.Create;
      Harddisks.Sorted := True;
      DriveNo := 0;
      NativeDir('\Device', Devices);
      for i := 0 to Devices.Count - 1 do
      begin
         DeviceName := '';

         if StartsWith(Devices[i], 'CdRom', Number) then
         begin
            DriveNo := StrToIntDef(Number, -1);
            if DriveNo >= 0 then
            begin
               DeviceName := '\Device\CdRom' + IntToStr(DriveNo);
            end;
         end
         else if StartsWith(Devices[i], 'Floppy', Number) then
         begin
            DriveNo := StrToIntDef(Number, -1);
            if DriveNo >= 0 then
            begin
               DeviceName := '\Device\Floppy' + IntToStr(DriveNo);
            end;
         end
         else if StartsWith(Devices[i], 'Harddisk', Number) then
         begin
            DriveNo := StrToIntDef(Number, -1);
            if DriveNo >= 0 then
            begin
               // scan the partitions...
               Harddisks.Add('\Device\Harddisk' + IntToStr(DriveNo));
            end;
         end;

         if Length(DeviceName) > 0 then
         begin
            Debug('Native DIR found ' + DeviceName, DebugHigh);

            if AddRemovable then
            begin
               NewPart := TPartition.Create;

               NewPart.DriveNo                  := DriveNo;
               NewPart.PartNo                   := PartNo;
               NewPart.FileName                 := DeviceName;
               NewPart.DeviceName               := DeviceName;
               NewPart.BytesPerSector           := 512;
               NewPart.MediaType                := Media_Type_RemovableMedia;
               NewPart.PartitionType            := PARTITION_LINUX; // we don;t know just yet...
               NewPart.RecognizedPartition      := True;
               NewPart.RewritePartition         := False;
               NewPart.StartingOffset.QuadPart  := 0;
               NewPart.PartitionLength.QuadPart := $FFFFFFFFFFFFFFFF;
               NewPart.HiddenSectors            := 0;

               Partitions.Add(NewPart);
            end;
         end;
      end;

      // do the hard disk partitions...
      for DriveNo := 0 to Harddisks.Count - 1 do
      begin
         Devices.Clear;
         NativeDir(Harddisks[DriveNo], Devices);
         for i := 0 to Devices.Count - 1 do
         begin
            if StartsWith(Devices[i], 'Partition', Number) then
            begin
               PartNo := StrToIntDef(Number, -1);
               DeviceName := Harddisks[DriveNo] + '\Partition' + IntToStr(PartNo);
               NTParts.Add(DeviceName);
            end;
         end;
      end;

      Devices.Free;
      Harddisks.Free;
   end;

   procedure HexDump(Offset : Int64; Length : Integer);
   var
      ReadStart  : _Large_Integer;
      Seek       : DWORD;
      Buffer     : String;
      S  : String;
      S2 : String;
      i  : Integer;
      Actual     : DWORD;
   begin
      ReadStart.QuadPart := Offset;
      Seek := SetFilePointer(h, ReadStart.LowPart, @ReadStart.HighPart, FILE_BEGIN);
      if Seek <> $FFFFFFFF then
      begin
         SetLength(Buffer, Length);
         if ReadFile2(h, PCHAR(Buffer), Length, Actual, nil) then
         begin
            Debug('Disk dump offset = ' + IntToStr(Offset) + ' size = ' + IntToStr(Length), DebugOff);
            Debug('actual read = ' + IntToStr(Actual), DebugOff);
            S := '0000 ';
            for i := 0 to Length - 1 do
            begin
               S := S + IntToHex(Ord(PChar(Buffer)[i]), 2);
               if not (ord(PChar(Buffer)[i]) in [0, 8, 9]) then
               begin
                  S2 := S2 + PChar(Buffer)[i];
               end
               else
               begin
                  S2 := S2 + ' ';
               end;
               S := S + ' ';
               if System.Length(S) >= 52 then
               begin
                  Debug(S + ' ' + S2, DebugOff);
                  S := IntToHex(i + 1, 4) + ' ';
                  s2 := '';
               end;
            end;
         end
         else
         begin
            Debug('Error reading file ' + IntToStr(GetLastError) + ' ' + SysErrorMessage(GetLastError), DebugHigh);
         end
      end
      else
      begin
         Debug('Seek failed error=' + IntToStr(GetLastError), DebugHigh);
      end;
   end;
begin
   DriveNo := 0;

   Done := False;


   NTParts := TStringList.Create;
   LVM2ClearPV;

   // this will add any removable devices
   // and populate NTParts with fixed places to look at
   FindNative(ScanFloppy, NTParts);

   //NTParts.Add('\??\H:\working\explore2fs\lvm2\lvm_debug\lvm_debug.hda3');

{   while not Done do
   begin
      PartNo := 0;
      while True do
      begin}

   i := 0;
//   for i := 0 to NTParts.Count - 1 do
   while i < NTParts.Count do
   begin
//         DeviceName := '\Device\Harddisk' + IntToStr(DriveNo) + '\Partition' + IntToStr(PartNo);
      DeviceName := NTParts[i];
         Debug('Looking at ' + DeviceName, DebugHigh);
         // try and open the file.
         h := NTCreateFile(PChar(DeviceName), GENERIC_READ, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);

         if h <> INVALID_HANDLE_VALUE then
         begin
            //EnableDASD(h);
            try
               NewPart := TPartition.Create;

               NewPart.DriveNo                  := DriveNo;
               NewPart.PartNo                   := PartNo;
               NewPart.FileName                 := DeviceName;
               NewPart.BytesPerSector           := 512;
               NewPart.MediaType                := Media_Type_FixedMedia;
               NewPart.PartitionType            := 0;
               NewPart.RecognizedPartition      := True;
               NewPart.RewritePartition         := False;
               if Assigned(NTParts.Objects[i]) then
               begin
               Debug('we have an offset...', DebugHigh);
                  NewPart.StartingOffset.QuadPart  := TOffset(NTParts.Objects[i]).Offset;
                  NewPart.PartitionLength.QuadPart := $FFFFFFFFFFFFFFFF;
               end
               else
               begin
               NewPart.StartingOffset.QuadPart  := 0;
               NewPart.PartitionLength.QuadPart := $FFFFFFFFFFFFFFFF;
               end;
               NewPart.HiddenSectors            := 0;

               TestVolumeRead(NewPart);
               if CheckForExt(NewPart) then
               begin
                  NewPart.PartitionType := PARTITION_LINUX;
                  Partitions.add(NewPart);
                  NewPart := nil;
               end
               else if CheckForReiser(NewPart) then
               begin
                  ReiserFSCount := ReiserFSCount + 1;
               end
               else if CheckForLVM(NewPart) then
               begin
                  Debug('Adding ' + DeviceName + ' to PV list ' + IntToStr(NewPart.StartingOffset.QuadPart), DebugLow);
                  LVM2AddPV(DeviceName, NewPart.StartingOffset.QuadPart);
               end;

               if Assigned(NewPart) then
               begin
                  // not used...
                  NewPart.Free;
               end;
            finally
               CloseHandle(h);
            end;
         end
         else
         begin
            ErrorNo := Native.GetLastError;
            if ErrorNo = ERROR_FILE_NOT_FOUND then
            begin
               if PartNo = 0 then
               begin
                  Done := True;
               end;
               break;
            end
            else if ErrorNo = ERROR_PATH_NOT_FOUND then
            begin
               Done := True;
               break;
            end
            else if ErrorNo = 5 then
            begin
               MessageDlg('This program requires Administrator privilages to read harddisks', mtError, [mbOK], 0);
               break;
            end
            else if ErrorNo = 32 then
            begin
               // in use by something else (probably mounted)
               // ignore
            end
            else
            begin
               Debug('GetLastError = (' + IntToStr(ErrorNo) + ') ' + SysErrorMessage(ErrorNo), DebugLow);
            end;
         end;
         PartNo := PartNo + 1;
//      end;
//      DriveNo := DriveNo + 1;
      Inc(i);
   end;

//   LVM2AddPV('\??\f:\lvm2.info');
   Debug('Device scan done', DebugHigh);
   LVMBuildLV(Partitions);
   // now check any partitions which might have been added by LVM
   for i := 0 to Partitions.Count - 1 do
   begin
      NewPart := TPartition(Partitions[i]);
      if NewPart.PartitionType = 0 then
      begin
         h := NTCreateFile(PChar(NewPart.FileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);

         if h <> INVALID_HANDLE_VALUE then
         begin
            if CheckForExt(NewPart) then
            begin
               NewPart.PartitionType := PARTITION_LINUX;
            end
            else
            begin
               if CheckForReiser(NewPart) then
               begin
                  ReiserFSCount := ReiserFSCount + 1;
               end;
               // delete it cos it aint ext
               Partitions[i] := nil;
               NewPart.Free;
            end;
            CloseHandle(h);
         end;
      end;
   end;
end;

function TPartitionList.MediaPresent(h : THandle) : Boolean;
var
   Length : DWord;
begin
   Debug('Looking for media...', DebugHigh);
   Result := DeviceIOControl(h, CtlCode(IOCTL_STORAGE_BASE, $200, METHOD_BUFFERED, FILE_READ_ACCESS),
       nil, 0, nil, 0, Length, nil)
//CTL_CODE(IOCTL_STORAGE_BASE, 0x0200, METHOD_BUFFERED, FILE_READ_ACCESS)
end;

procedure TPartitionList.Dump;
var
   i : Integer;
begin
   for i := 0 to Partitions.Count - 1 do
   begin
      TPartition(Partitions[i]).Dump;
   end;
end;

procedure TPartition.Dump;
begin
   Debug('Partition number ' + IntToStr(PartNo) + ' on drive ' + IntToStr(DriveNo), DebugMedium);
   Debug(FileName, DebugMedium);

   Debug('Media type ' + IntToStr(MediaType), DebugMedium);
   Debug('Cylinders        : ' + IntToStr(Cylinders.QuadPart), DebugMedium);
   Debug('Tracks / Cylinder : ' + IntToStr(TracksPerCylinder), DebugMedium);
   Debug('Sectors / Track   : ' + IntToStr(SectorsPerTrack), DebugMedium);
   Debug('Bytes / Sector    : ' + IntToStr(BytesPerSector), DebugMedium);

   Debug('Partition:', DebugMedium);

   Debug('Type       : ' + IntToHex(PartitionType, 4), DebugMedium);
   Debug('Bootable   : ' + YesNo(BootIndicator), DebugMedium);
   Debug('Recognised : ' + YesNo(RecognizedPartition), DebugMedium);
   Debug('Rewrite    : ' + YesNo(RewritePartition), DebugMedium);

   Debug('Starting Offset : ' + IntToStr(StartingOffset.QuadPart), DebugMedium);

   Debug('Partition Length : ' + IntToStr(PartitionLength.QuadPart), DebugMedium);
   Debug('Hidden sectors : ' + IntToStr(HiddenSectors), DebugMedium);
   Debug('', DebugMedium);
   Debug('', DebugMedium);
end;

////////////////////////////////////
////////////////////////////////////
// Windows 95 scan
// This does a Low Level scan (reads partition table from disk);
// This is the same as LL_Scan, but it uses the DiskIO unit

procedure TPartitionList.Win95_Scan(PartType : Integer; UseEI13 : Boolean);
var
   DriveNo  : Integer;
   PartNo   : Integer;
   ExtendedStart : ULong;
   Disk     : T95Disk;

   procedure AddFloppy(Disk : T95Disk);
   var
      Size     : DWORD;
      NewPart  : TPartition;
   begin
      NewPart := TPartition.Create;
      Partitions.add(NewPart);
      NewPart.DriveNo   := DriveNo;
      NewPart.FileName  := '';

      NewPart.Cylinders.QuadPart := Disk.GetGeometry.Cylinder + 1;
      NewPart.TracksPerCylinder  := Disk.GetGeometry.Head + 1;
      NewPart.SectorsPerTrack    := Disk.GetGeometry.Sector;
      NewPart.BytesPerSector     := 512;
      NewPart.MediaType          := Media_Type_RemovableMedia;

      Size := Disk.SectorCount * 512;

      NewPart.PartitionType            := PartType;
      NewPart.BootIndicator            := False;
      NewPart.RecognizedPartition      := True;
      NewPart.RewritePartition         := False;
      NewPart.PartNo                   := 0;
      NewPart.StartingOffset.QuadPart  := 0;
      NewPart.PartitionLength.QuadPart := Size;
      NewPart.PartitionLength.QuadPart := NewPart.PartitionLength.QuadPart * 512;
      NewPart.HiddenSectors            := 0;
   end;

   function CheckForReiser(Disk : T95Disk; Sector : ULong) : Boolean;
   var
      Buffer     : String;
      Magic      : String;
   begin
      Result := False;
      SetLength(Buffer, 512);
      if Disk.ReadSector(Sector + 128, PChar(Buffer), 1) then
      begin
         // offset 52 should contain 'ReIsEr2Fs'
         Magic := Copy(Buffer, 53, 9);
         Debug('Magic = ' + Magic, DebugHigh);
         if Magic = 'ReIsEr2Fs' then
         begin
            Debug('ReiserFS Found. magic = ' + Magic, DebugOff);
            Result := True;
         end;
      end;
   end;


   function ProcessSector(Disk : T95Disk; Sector : ULong; Depth : Integer) : Boolean;
   var
      Size     : DWORD;
      Buffer   : TDiskTable;
      i        : Integer;
      NewPart  : TPartition;
      Superblock : ext2_super_block;

   begin
      Result := True;
      Size := sizeof(Buffer);
      Debug('Looking for table at sector ' + IntToStr(Sector), DebugMedium);

      if not Disk.ReadSector(Sector, @Buffer, size div 512) then
         begin
            Result := False;
            exit;
         end;
         Debug('Signature is ' + IntToHex(Buffer.Signature, 4), DebugMedium);
         if (Buffer.Signature = $AA55) then
         begin
            for i := 1 to 4 do
            begin
               Debug('Processing partitions', DebugMedium);
               if Buffer.Table[i].sys_ind <> 0 then
               begin
                  Debug('Partition           ' + IntToStr(i) + ' /dev/hdx' + IntToStr(PartNo), DebugMedium);
                  Debug(' boot indicator      ' + IntToHex(Buffer.Table[i].boot_ind,2), DebugHigh);
                  Debug(' starting head       ' + IntToStr(Buffer.Table[i].head), DebugHigh);
                  Debug(' starting sector     ' + IntToStr(Buffer.Table[i].sector), DebugHigh);
                  Debug(' starting cylinder   ' + IntToStr(Buffer.Table[i].cyl), DebugHigh);
                  Debug(' partition type      ' + IntToHex(Buffer.Table[i].sys_ind,2), DebugMedium);
                  case Buffer.Table[i].sys_ind of
                     PARTITION_EXTENDED         : Debug('Extended... (dos)', DebugMedium);
                     PARTITION_EXTENDED_LINUX   : Debug('Extended... (linux)', DebugMedium);
                     PARTITION_EXTENDED_WIN98   : Debug('Extended... (win98)', DebugMedium);
                     PARTITION_FAT_16           : Debug('FAT 16', DebugMedium);
                     PARTITION_HUGE             : Debug('VFAT', DebugMedium);
                     PARTITION_IFS              : Debug('NTFS', DebugMedium);
                     PARTITION_LINUX            : Debug('Linux', DebugMedium);
                     PARTITION_LINUX_SWAP       : Debug('Linux Swap', DebugMedium);
                  else
                     Debug('Unknown', DebugMedium);
                  end;
                  Debug('end head            ' + IntToStr(Buffer.Table[i].end_head), DebugHigh);
                  Debug('end sector          ' + IntToStr(Buffer.Table[i].end_sector), DebugHigh);
                  Debug('end cylinder        ' + IntToStr(Buffer.Table[i].end_cyl), DebugHigh);
                  Debug(' starting sector     ' + IntToStr(Buffer.Table[i].start_sect), DebugMedium);
                  Debug(' number of sectors   ' + IntToStr(Buffer.Table[i].nr_sects), DebugMedium);
                  Debug('', DebugMedium);

                  NewPart := nil;
                  if Disk.ReadSector(Buffer.Table[i].start_sect + 2, @superblock, sizeof(superblock) div 512) then
                  begin
                     if superblock.s_magic = EXT2_SUPER_MAGIC then
                     begin
                        Debug('EXT2 Found. magic = 0x' + IntToHex(superblock.s_magic, 4), DebugOff);

                        if (Buffer.Table[i].sys_ind = PartType) or (PartType = 0) then
                        begin
                           NewPart := TPartition.Create;
                           Partitions.add(NewPart);
                           NewPart.DriveNo   := DriveNo;// - $80;
                           NewPart.FileName  := '';

                           NewPart.Cylinders.QuadPart := Disk.GetGeometry.Cylinder + 1;
                           NewPart.TracksPerCylinder  := Disk.GetGeometry.Head + 1;
                           NewPart.SectorsPerTrack    := Disk.GetGeometry.Sector;
                           NewPart.BytesPerSector     := 512;
                           NewPart.MediaType          := Media_Type_FixedMedia;

                           NewPart.PartitionType            := Buffer.Table[i].sys_ind;
                           NewPart.BootIndicator            := (Buffer.Table[i].boot_ind = $80);
                           NewPart.RecognizedPartition      := True;
                           NewPart.RewritePartition         := False;
                           NewPart.PartNo                   := PartNo;
                           NewPart.StartingOffset.QuadPart  := (Buffer.Table[i].start_sect + Sector);// Offset from start of partition
   //                        NewPart.StartingOffset.QuadPart  := NewPart.StartingOffset.QuadPart * Geometry.BytesPerSector;
                           NewPart.PartitionLength.QuadPart := Buffer.Table[i].nr_sects;
   //                        NewPart.PartitionLength.QuadPart := NewPart.PartitionLength.QuadPart * Geometry.BytesPerSector;
                           NewPart.HiddenSectors            := 0;

                        end;
                     end;
                     if (Buffer.Table[i].sys_ind in EXTENDED_PARTITIONS) and (PartNo > 4) then
                     begin
                        Dec(PartNo); // We do not want to assign a number to extend partitions after the 1st (I don't know why?, it is what Linux does)
                     end;
                  end;

                  if NewPart = nil then
                  begin
                     if CheckForReiser(Disk, Buffer.Table[i].start_sect) then
                     begin
                        ReiserFSCount := ReiserFSCount + 1;
                     end;
                  end;

                  Inc(PartNo);
               end;
            end;
            for i := 1 to 4 do
            begin
               if Buffer.Table[i].sys_ind in EXTENDED_PARTITIONS then
               begin
                  Debug('Processing extended partition....', DebugMedium);
                  if Depth < 32 then
                  begin
                     if PartNo < 5 then
                     begin
                        PartNo := 5;
                     end;
                     if Depth = 1 then
                     begin
                        ExtendedStart := Sector;
                     end;
                     // There is a new partition table at offset start_sect...
                     if Depth >= 2 then
                     begin
                     if not ProcessSector(Disk, ExtendedStart + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Debug('Did not seem to be a valid extended partition', DebugMedium);
                        end;
                     end
                     else
                     begin
                        if not ProcessSector(Disk, Sector + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Debug('Did not seem to be a valid extended partition', DebugMedium);
                        end;
                     end;
                  end
                  else
                  begin
                     Debug('More than 32 partitions detected! This is most likely a bug!', DebugOff);
                  end;
               end;
            end;
         end
         else
         begin
            Result := False;
         end;
      end;
begin
   Disk := T95Disk.Create(UseEI13);
   try
      DriveNo := $80;

      while true do
      begin
         Debug('Scanning disk ' + IntToHex(DriveNo, 2), DebugMedium);
         if not Disk.SetDisk(DriveNo) then
         begin
               // no more physical drives found...
               break;
         end
         else // we have the handle, lets look at the partitions....
         begin
            // first see if the media is present...
               // Time to read the table
               PartNo := 1;
               ExtendedStart := 0;
               ProcessSector(Disk, 0, 0);
            Inc(DriveNo);
         end;
      end;

      ////////////////////////////////////////////
      ////////////////////////////////////////////
      ////////////////////////////////////////////
      ////////////////////////////////////////////

      if ScanFloppy then
      begin
         DriveNo := 0;

         while true do
         begin
            Debug('Scanning ' + IntToHex(DriveNo, 2), DebugMedium);
            if not Disk.SetDisk(DriveNo) then
            begin
               break;
            end
            else // we have the handle, lets look at the partitions....
            try
               // Time to read the table
               PartNo := 1;
               ExtendedStart := 0;
               ProcessSector(Disk, 0, 0);
               Debug('PartNo = ' + IntToStr(PartNo), DebugMedium);
//               if PartNo = 1 then
               begin
                  // no Partitons found, should we add the whole disk?
                  AddFloppy(Disk);
               end;
            finally
               Inc(DriveNo);
            end;
         end;
      end;
   finally
      Disk.Free;
   end;
end;


end.
